/**************************************************************************
    bme280.c

    pressuer, temperature, humidity multi sensor 

    2022.2.3  (c) Suwa-Koubou

***************************************************************************/

#include <stdio.h>
//#include <stdlib.h>
#include "HardwareProfile.h"
#include "pic_delay.h"
#include "bme280.h"
#include "tskchg.h"

#include "debug.h"

/*--------------------------------------------------------------------------
    definition
--------------------------------------------------------------------------*/

// I2C data transfer mode bit
#define MW          0   // I2C master write mode
#define MR          1   // I2C master read mode

// register map
#define CALIB00     0x88
#define CALIB25     0xA1
#define ID          0xD0
#define RESET       0xE0
#define CALIB26     0xE1
#define CALIB41     0xF0
#define CTRL_HUM    0xF2
#define STATUS      0xF3
#define CTRL_MEAS   0xF4
#define CONFIG      0xF5
#define PRESS_MSB   0xF7
#define PRESS_LSB   0xF8
#define PRESS_XLSB  0xF9
#define TEMP_MSB    0xFA
#define TEMP_LSB    0xFB
#define TEMP_XLSB   0xFC
#define HUM_MSB     0xFD
#define HUM_LSB     0xFE


/*--------------------------------------------------------------------------
    gloval variable
--------------------------------------------------------------------------*/

// temperature adjust data
static unsigned short dig_T1;
short          dig_T2;
short          dig_T3;
 
// humidity adjust data
unsigned char  dig_H1;
short          dig_H2;
unsigned char  dig_H3;
short          dig_H4;
short          dig_H5;
char           dig_H6;
 
// pressuer adjust data
unsigned short dig_P1;
short          dig_P2;
short          dig_P3;
short          dig_P4;
short          dig_P5;
short          dig_P6;
short          dig_P7;
short          dig_P8;
short          dig_P9;

static int     t_fine;

/*--------------------------------------------------------------------------
    function prototype
--------------------------------------------------------------------------*/

static unsigned int bme280_compensate_P_int32(int adc_P);
static int          bme280_compensate_T_int32(int adc_T);
static unsigned int bme280_compensate_H_int32(int adc_H);

void bme280_test(void)
{
    char  data;
    int  press, temp, hum;
    char pbuf[32];

    data=bme280Read(ID);
    putstr("BME280_ID=0x");
    puthex(data);
    putcrlf();

    for(;;){
        bme280ReadData(&press, &temp, &hum);
        sprintf(pbuf, "press=%d\r\n", press); putstr(pbuf);
        sprintf(pbuf, "temp=%d\r\n", temp); putstr(pbuf);
        sprintf(pbuf, "hum=%d\r\n", hum); putstr(pbuf);
        delay_ms(30000); // wait 10sec
    }
}

 
/*==========================================================================
    enviroment sensor BME280 initialize
==========================================================================*/

/*--------------------------------------------------------------------------
    BME280 initialize
--------------------------------------------------------------------------*/

void    bme280Init(void)
{
    char config_val, ctrl_meas_val, ctrl_hum_val;
    unsigned char dac[26];

    delay_ms(5);
 
//putstr("bme280Init....\r\n");

    bme280Write(RESET, 0xB6);     // sensor soft reset
    delay_ms(5);
   
    config_val    = 0b00000000;   // t_sb=0.5ms, filter=off, spi3w=disable
    ctrl_meas_val = 0b00100100;   // osrs_t=x1,osrs_p=x1,mode=sleep
    ctrl_hum_val  = 0b00000001;   // osrs_h=x1

    bme280Write(CONFIG, config_val);
    bme280Write(CTRL_MEAS, ctrl_meas_val);
    bme280Write(CTRL_HUM, ctrl_hum_val);

    delay_ms(5);

//putstr("CONFIG=");     puthex(bme280Read(CONFIG));    putcrlf();
//putstr("CTRL_MEAS=");  puthex(bme280Read(CTRL_MEAS)); putcrlf();
//putstr("CTRL_HUM=");   puthex(bme280Read(CTRL_HUM));  putcrlf();

    // read adjust data
    bme280ReadBlock(CALIB00, (char *)dac, 26); // read calib00 - calib25

    dig_T1 = ((unsigned short)((dac[1] << 8) | dac[0]));
    dig_T2 = ((short)((dac[3] << 8) | dac[2]));
    dig_T3 = ((short)((dac[5] << 8) | dac[4]));

//putstr("dig_T1=0x"); putword(dig_T1); putcrlf();
//putstr("dig_T2=0x"); putword(dig_T2); putcrlf();
//putstr("dig_T3=0x"); putword(dig_T3); putcrlf();


    dig_P1 = ((unsigned short)((dac[7] << 8) | dac[6]));
    dig_P2 = ((short)((dac[9] << 8) | dac[8]));
    dig_P3 = ((short)((dac[11] << 8) | dac[10]));
    dig_P4 = ((short)((dac[13] << 8) | dac[12]));
    dig_P5 = ((short)((dac[15] << 8) | dac[14]));
    dig_P6 = ((short)((dac[17] << 8) | dac[16]));
    dig_P7 = ((short)((dac[19] << 8) | dac[18]));
    dig_P8 = ((short)((dac[21] << 8) | dac[20]));
    dig_P9 = ((short)((dac[23] << 8) | dac[22]));

//putstr("dig_P1=0x"); putword(dig_P1); putcrlf();
//putstr("dig_P2=0x"); putword(dig_P2); putcrlf();
//putstr("dig_P3=0x"); putword(dig_P3); putcrlf();
//putstr("dig_P4=0x"); putword(dig_P4); putcrlf();
//putstr("dig_P5=0x"); putword(dig_P5); putcrlf();
//putstr("dig_P6=0x"); putword(dig_P6); putcrlf();
//putstr("dig_P7=0x"); putword(dig_P7); putcrlf();
//putstr("dig_P8=0x"); putword(dig_P8); putcrlf();
//putstr("dig_P9=0x"); putword(dig_P9); putcrlf();

    dig_H1 = ((unsigned char)(dac[25]));

//putstr("dig_H1=0x"); putword(dig_H1); putcrlf();

    bme280ReadBlock(CALIB26, (char *)dac, 7); // read calib26 - calib32

//{
// int n;
// putstr("dac[7]=");
// for(n=0;n<7;n++){
//  puthex(dac[n]); putch(' ');
// }
// putcrlf();
//}

    dig_H2 = ((short)((dac[1] << 8) | dac[0]));
    dig_H3 = ((unsigned char)(dac[2]));
    dig_H4 = ((short)((dac[3] << 4) + (dac[4] & 0x0F)));
    dig_H5 = ((short)((dac[5] << 4) + ((dac[4] >> 4) & 0x0F)));
    dig_H6 = ((char)dac[6]);

//putstr("dig_H2=0x"); putword(dig_H2); putcrlf();
//putstr("dig_H3=0x"); putword(dig_H3); putcrlf();
//putstr("dig_H4=0x"); putword(dig_H4); putcrlf();
//putstr("dig_H5=0x"); putword(dig_H5); putcrlf();
//putstr("dig_H6=0x"); putword(dig_H6); putcrlf();

    delay_ms(10); //wait 10msec
}

/*--------------------------------------------------------------------------
    bme280ReadData(int *press, int *temp, int *hum)
--------------------------------------------------------------------------*/

extern int  readTempFromLM61BIZ(void);   // 2022.4.11
char pbuf[32];

void bme280ReadData(int *press, int *temp, int *hum)
{
    unsigned char dac[8];
    int  adc_P, adc_T, adc_H;

    unsigned int pres_cal, humi_cal;
    int          temp_cal;

    // force measurement
    bme280Write(CTRL_MEAS, 0x25);  // osrs_t=x1,osrs_p=x1,mode=force
    //delay_ms(3);
    tsk_delay(10); // wait 10msec

    // get data PRESS_MSB to HUM_LSB
    bme280ReadBlock(PRESS_MSB, (char *)dac, 8);


    adc_P = ((unsigned int)dac[0] << 12) | ((unsigned int)dac[1] << 4) | ((dac[2] >> 4) & 0x0F);
    adc_T = ((unsigned int)dac[3] << 12) | ((unsigned int)dac[4] << 4) | ((dac[5] >> 4) & 0x0F);
    adc_H = ((unsigned int)dac[6] << 8) | ((unsigned int)dac[7]);

//{
//   int n;
//   putstr("dac[8]=");
//   for(n=0;n<8;n++){
//       puthex(dac[n]); putch(' ');
//   }
//   putcrlf();

//   putstr("adc_P=0x"); putdword(adc_P); putcrlf();
//   putstr("adc_T=0x"); putdword(adc_T); putcrlf();
//   putstr("adc_H=0x"); putdword(adc_H); putcrlf();

//}

    temp_cal = bme280_compensate_T_int32(adc_T); //xf[^␳vZ
    humi_cal = bme280_compensate_H_int32(adc_H); //xf[^␳vZ
    pres_cal = bme280_compensate_P_int32(adc_P); //Cf[^␳vZ

    *press = pres_cal/100;     //Cf[^ۂ̒lɌvZ
    *temp =  temp_cal/10;      //xf[^ۂ̒lɌvZ
    *hum =   humi_cal*10/1024; //xf[^ۂ̒lɌvZ

sprintf(pbuf,"BME280=%d\r\n", *temp); putstr(pbuf);

    // 2022.4.11
    // Temperature changed from BME280 to LM61BIZ
    *temp = readTempFromLM61BIZ();

sprintf(pbuf,"LM61B=%d\r\n", *temp); putstr(pbuf);

}


/*--------------------------------------------------------------------------
    temp adjust
--------------------------------------------------------------------------*/

static int bme280_compensate_T_int32(int adc_T)
{
    int var1, var2, T;
    var1  = ((((adc_T>>3) - ((int)dig_T1<<1))) * ((int)dig_T2)) >> 11;
    var2  = (((((adc_T>>4) - ((int)dig_T1)) * ((adc_T>>4) - ((int)dig_T1))) >> 12) * ((int)dig_T3)) >> 14;
    t_fine = var1 + var2;
    T  = (t_fine * 5 + 128) >> 8;
    return T;
}

/*--------------------------------------------------------------------------
    humidity adjust
--------------------------------------------------------------------------*/

static unsigned int bme280_compensate_H_int32(int adc_H)
{
    int v_x1_u32r;

    v_x1_u32r = (t_fine - ((int)76800)); 
    v_x1_u32r = (((((adc_H << 14) - (((int)dig_H4) << 20) - (((int)dig_H5) * v_x1_u32r)) + ((int)16384)) >> 15) *
                 (((((((v_x1_u32r * ((int)dig_H6)) >> 10) * (((v_x1_u32r * ((int)dig_H3)) >> 11) + ((int)32768))) >> 10) + ((int)2097152)) * ((int)dig_H2) + 8192) >> 14) );
    v_x1_u32r = (v_x1_u32r - (((((v_x1_u32r >> 15) * (v_x1_u32r >> 15)) >> 7) * ((int)dig_H1)) >> 4));
    v_x1_u32r = (v_x1_u32r < 0 ? 0 : v_x1_u32r);
    v_x1_u32r = (v_x1_u32r > 419430400 ? 419430400 : v_x1_u32r);

    return (unsigned int)(v_x1_u32r>>12);
}

/*--------------------------------------------------------------------------
    pressuer adjust
--------------------------------------------------------------------------*/

static unsigned int bme280_compensate_P_int32(int adc_P)
{
    int var1, var2;
    unsigned int p;

    var1 = (((int)t_fine)>>1) - (int)64000;
    var2 = (((var1>>2) * (var1>>2)) >> 11 ) * ((int)dig_P6);
    var2 = var2 + ((var1*((int)dig_P5))<<1);
    var2 = (var2>>2)+(((int)dig_P4)<<16);
    var1 = (((dig_P3 * (((var1>>2) * (var1>>2)) >> 13 )) >> 3) + ((((int)dig_P2) * var1)>>1))>>18;
    var1 =((((32768+var1))*((int)dig_P1))>>15);
    if (var1 == 0){
        return 0; // avoid exception caused by division by zero
    }
    p = (((unsigned int)(((int)1048576)-adc_P)-(var2>>12)))*3125;
    if (p < 0x80000000){
        p = (p << 1) / ((unsigned int)var1);
    }else{
        p = (p / (unsigned int)var1) * 2;
    }
    var1 = (((int)dig_P9) * ((int)(((p>>3) * (p>>3))>>13)))>>12;
    var2 = (((int)(p>>2)) * ((int)dig_P8))>>13;
    p = (unsigned int)((int)p + ((var1 + var2 + dig_P7) >> 4));

    return p;
}


/*--------------------------------------------------------------------------
    BME280 write byte
--------------------------------------------------------------------------*/

void    bme280Write(char addr, char data)
{
    i2cOpenStart();
    i2cSend(BMEDEVADR + MW);
    i2cSend(addr);
    i2cSend(data);
    i2cStopClose();
}   


/*--------------------------------------------------------------------------
     BME280 read byte
--------------------------------------------------------------------------*/

int  bme280Read(char addr)
{
    int val;

    i2cOpenStart();
    i2cSend(BMEDEVADR + MW);
    i2cSend(addr);
    // i2cReStart();
    i2cStopClose();
    i2cOpenStart();
    i2cSend(BMEDEVADR + MR);
    val = i2cRec();
    i2cNAckSend();
    i2cStopClose();

    return val & 0xff;
}   


/*--------------------------------------------------------------------------
     BME280 read block
--------------------------------------------------------------------------*/

void  bme280ReadBlock(char addr, char data[], int block_size)
{
    int  i;

    i2cOpenStart();
    i2cSend(BMEDEVADR + MW);
    i2cSend(addr);
    // i2cReStart();
    i2cStopClose();
    i2cOpenStart();
    i2cSend(BMEDEVADR + MR);
    for(i = 0; i < block_size; i++){
        data[i] = i2cRec();
        if(i < block_size-1){
            i2cAckSend();
        }else{
            i2cNAckSend();
        }
    }
    i2cStopClose();

}



#ifdef  I2C_HANDLING

/*==========================================================================
    I2C Handling module
==========================================================================*/
#define TIME_OUT    0xffff
static unsigned short tw;

#define  BRG_40M_400KHZ     0x02c
#define  BRG_40M_100KHZ     0x0c2

/*--------------------------------------------------------------------------
    I2C master Open and Start
--------------------------------------------------------------------------*/

void    i2cOpenStart(void)
{
    I2C1BRG = BRG_40M_400KHZ;
    //I2C1BRG = BRG_40M_100KHZ;

    // enabled I2C
    I2C1CONbits.I2CEN = 1;

    I2C1CONbits.I2CEN = 1;
    delay_us(5);

    // START 
    I2C1CONbits.SEN = 1;
    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2C1CONbits.SEN) break;
    }
}


/*--------------------------------------------------------------------------
    I2C master ReStart
--------------------------------------------------------------------------*/

void    i2cReStart(void)
{
    // RESTART 
    I2C1CONbits.RSEN = 1;
    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2C1CONbits.RSEN) break;
    }
}


/*--------------------------------------------------------------------------
    I2C master Stop and Close
--------------------------------------------------------------------------*/

void    i2cStopClose(void)
{
    // STOP 
    I2C1CONbits.PEN = 1;
    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2C1CONbits.PEN)break;
    }
    delay_us(5);

    // disabled I2C
    I2C1CONbits.I2CEN = 0;
}


/*--------------------------------------------------------------------------
    I2C master data send
--------------------------------------------------------------------------*/

int i2cSend(unsigned char data)
{
    I2C1TRN = data;

    if(I2C1STATbits.IWCOL){ // If write collision occurs,return -2
        return -2;
    }

    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2C1STATbits.TRSTAT) break;  // wait until write cycle is complete 
    }

    if (I2C1STATbits.ACKSTAT){  // test for ACK condition received
        return -1;  // NACK received
    }

    return 0;   // complete with no error.
}


/*--------------------------------------------------------------------------
    I2C master data receive
--------------------------------------------------------------------------*/

unsigned char   i2cRec(void)
{
    // enabled Receive
    I2C1CONbits.RCEN = 1;

    // wait until receive complete and data stored to I2CRCV register
    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2C1CONbits.RCEN && I2C1STATbits.RBF) break;
    }

    // clear overflow bit
    I2C1STATbits.I2COV = 0;
    
    return (unsigned char)I2C1RCV;
}


/*--------------------------------------------------------------------------
    I2C master Ack send
--------------------------------------------------------------------------*/

void    i2cAckSend(void)
{
    // Ack Send
    I2C1CONbits.ACKDT = 0;
    I2C1CONbits.ACKEN = 1;

    // wait until ack cycle complete
    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2C1CONbits.ACKEN)break;
    }
}


/*--------------------------------------------------------------------------
    I2C master NotAck send
--------------------------------------------------------------------------*/

void    i2cNAckSend(void)
{
    // NAck Send
    I2C1CONbits.ACKDT = 1;
    I2C1CONbits.ACKEN = 1;

    // wait until ack cycle complete
    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2C1CONbits.ACKEN) break;
    }
}


#endif  // I2C_HANDLING

/*** end of bme280c ******************************************************/
